/*
 * samygodca_inj.cpp
 * Copyright (C) 2010 Andreas Fiedler <smartsmurf@users.sourceforge.net>
 *
 * This file is part of samygodca, a free DTS Coherent Acoustics
 * stream decoder plugin for SamyGO extension of Samsung DTV.
 *
 * samygodca is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * samygodca is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "samygodca_inj.hpp"
#include "Logging.h"
#include <arpa/inet.h>

/* ffmpeg context */
static AVCodecContext  *pCodecCtx;
static AVCodec         *pCodec;
static AVPacket        avpacket; 
DLL_LOCAL struct ffmpeg_interface_t ffmpeg;

/* libdca context */
#ifdef __HAVE_LIBDCA__
static dca_state_t * libdca_state;
DLL_LOCAL struct libdca_interface_t libdca;

/* might be set externally */
static int samygodca_use_libdca = 1;
#endif

/* uncomment to dump decoded PCM data into file */
// #define DUMP_RAWPCM 1

#if DUMP_RAWPCM
#define RAWPCM_DUMPFILE "/dtv/pcmdump"
static FILE * file_pcmdump; 
#endif

static const unsigned char audiotransfer_sync_header[] = {0x7A,0x7B,0x0C,0x9D,0x1E,0x6F,0x3A,0x4B,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}; 

/* temporary storage */
static unsigned char audio_header[32];
static int audio_header_size;

enum CodecID soft_codec_id;

static unsigned char * framebuf;
static short * samples;
static int total_samples;
static int last_pts;

/* replacement for CParsingMedia::PrepareParsing(void) */
void CParsingMedia_PrepareParsing(CParsingMedia * Instance)
{
	PRINTF("%s In", __FUNCTION__ );
	//g_u32PrintLevel |= 0x300000;
	//Instance->PrepareParsing();
	Instance->PrepareParsingExtFfmpeg();
}

/* replacement for CMovieCore::SelectAudio(...) */
/*
int CMovieCore_SelectAudio(CMovieCore * Instance, Multimedia::_tMediaInfoContext * media_info_context, unsigned int audio_stream_index)
{
	PRINTF("%s In", __FUNCTION__ );
	return Instance->SelectAudio(media_info_context, audio_stream_index);
}
*/

/* replacement for CParsingMedia::GetMediaTypeInfo(char  const* file_name, Multimedia::_tMediaInfoContext * media_info_context ) */
int CParsingMedia_GetMediaTypeInfo(CParsingMedia * Instance, char  const* file_name, Multimedia::_tMediaInfoContext * media_info_context )
{
	int retval;
	PRINTF("%s In", __FUNCTION__ );
	retval = Instance->GetMediaTypeInfo(file_name, media_info_context);
	PRINTF("%s Out", __FUNCTION__ );
	return retval;
}

/* replacement for CheckSupportMovie(CMovieTDIF::MovieSettings) */
int CheckSupportMovie_DTS( CMovieTDIF::MovieSettings settings )
{
	return CheckSupportMovie( settings );
}

static void change_endian16( short * OutBuf, const short * InBuf, const int words )
{
	const short *in_end = InBuf + words;

	for (; InBuf < in_end; InBuf++)
		*OutBuf++ = htons(*InBuf);
}


#ifdef __HAVE_LIBDCA__
/* use external LIBDCA libraries */
static void PrepareLibDCA( struct libdca_interface_t * dca, const char * lib_path )
{
	char module_path[256];

	PRINTF("[%s] In", __FUNCTION__ );

	if( dca->initialised ){
		PRINTF("[%s] libdca already initialised.", __FUNCTION__ );
		return;
	}
 	snprintf(module_path, sizeof(module_path), "%s/%s", lib_path, "libdca.so"); 
	dca->hLib_libdca	= dlopen( module_path, RTLD_NOW | RTLD_LOCAL | RTLD_DEEPBIND );
	if( dca->hLib_libdca == NULL ){
		PRINTF("[%s dlopen error]  %s", module_path, dlerror() );
		goto exit_prep_libdca;
	}
	dca->dca_init = (dca_state_t *(*)(uint32_t))dlsym( dca->hLib_libdca, "dca_init" );
	if( dca->dca_init == NULL ){
		PRINTF("[dca->dca_init dlsym error]");
	}
	dca->dca_syncinfo = (int (*)(dca_state_t *, uint8_t *, int *, int *, int *, int *))dlsym( dca->hLib_libdca, "dca_syncinfo" );
	if( dca->dca_syncinfo == NULL ){
		PRINTF("[dca->dca_syncinfo dlsym error]");
	}
	dca->dca_frame = (int (*)(dca_state_t *, uint8_t *, int *, level_t *, sample_t))dlsym( dca->hLib_libdca, "dca_frame" );
	if( dca->dca_frame == NULL ){
		PRINTF("[dca->dca_frame dlsym error]");
	}
	dca->dca_dynrng = (void (*)(dca_state_t *, level_t (* ) (level_t, void *), void *))dlsym( dca->hLib_libdca, "dca_dynrng" );
	if( dca->dca_dynrng == NULL ){
		PRINTF("[dca->dca_dynrng dlsym error]");
	}
	dca->dca_blocks_num = (int (*)(dca_state_t *))dlsym( dca->hLib_libdca, "dca_blocks_num" );
	if( dca->dca_blocks_num == NULL ){
		PRINTF("[dca->dca_blocks_num dlsym error]");
	}
	dca->dca_block = (int (*)(dca_state_t *))dlsym( dca->hLib_libdca, "dca_block" );
	if( dca->dca_block == NULL ){
		PRINTF("[dca->dca_block dlsym error]");
	}
	dca->dca_samples = (sample_t * (*)(dca_state_t *))dlsym( dca->hLib_libdca, "dca_samples" );
	if( dca->dca_samples == NULL ){
		PRINTF("[dca->dca_samples dlsym error]");
	}
	dca->dca_free = (void (*)(dca_state_t *))dlsym( dca->hLib_libdca, "dca_free" );
	if( dca->dca_free == NULL ){
		PRINTF("[dca->dca_free dlsym error]");
	}
	/* all OK */
	dca->initialised = 1;

exit_prep_libdca:
	PRINTF("[%s] Out", __FUNCTION__ );
	return;

}

static inline int16_t libdca_convert (int32_t i)
{
#ifdef LIBDCA_FIXED
    i >>= 15;
#else
    i -= 0x43c00000;
#endif
    return (i > 32767) ? 32767 : ((i < -32768) ? -32768 : i);
}

static void libdca_convert2s16_2 (void * _f, int16_t * s16)
{
    int i;
    int32_t * f = (int32_t *) _f;

    for (i = 0; i < 256; i++) {
		s16[2*i] = libdca_convert (f[i]);
		s16[2*i+1] = libdca_convert (f[i+256]);
    }
}

static int libdca_dcatopcm( short * pcm_out, unsigned char * dca_in, int dca_in_size, int big_endian )
{
    int sample_rate;
    int frame_length;
    int flags;
	int bit_rate;
	int length;
	level_t level;
	sample_t bias; 
	int i;
	int out_size = 0;
	int num_blocks;
	short * pcm_out_ptr = pcm_out;

	length = libdca.dca_syncinfo( libdca_state, dca_in, &flags, &sample_rate, &bit_rate, &frame_length );
	if (!length) {
		LOG("Error returned by dca_syncinfo()");
		PRINTHEX("DCA Frame data/32:", dca_in, 32 );
		return -1;
	}
	flags = DCA_STEREO;
#ifdef LIBDCA_FIXED
	level = (1 << 26);	/* see libao/audio_out_internal.h  */
	bias  = 0;			/* see libao/audio_out_internal.h  */
#else
	level = 1;			/* see libao/audio_out_internal.h  */
	bias  = 384;		/* see libao/audio_out_internal.h  */
#endif
	if( libdca.dca_frame( libdca_state, dca_in, &flags, &level, bias) ){
		LOG("Error returned by dca_frame()");
		PRINTHEX("DCA Frame data/32:", dca_in, 32 );
		return -1;
	}
	/* disable dynrng */
	libdca.dca_dynrng( libdca_state, NULL, NULL );

	num_blocks = libdca.dca_blocks_num(libdca_state);
	// PRINTF("Number of dca blocks = %i", num_blocks );
	// PRINTF("dca flags = %i, level = 0x%08X, bias = %i", flags, level, bias );
	for (i = 0; i < num_blocks; i++) {
		if( libdca.dca_block(libdca_state) ){
			LOG("Error returned by dca_block()");
			return 0;
		}
		// libdca_sample_to_short( pcm_out_ptr, libdca.dca_samples(libdca_state), 256 );
		libdca_convert2s16_2( libdca.dca_samples(libdca_state), pcm_out_ptr );
		pcm_out_ptr += 512;
		out_size += 1024;
	}

	/* increment number of samples */
	total_samples += (out_size >> 1);

	if( big_endian )
		change_endian16( pcm_out, pcm_out, out_size >> 1 );	

#if DUMP_RAWPCM
		fwrite( (void *)pcm_out, 1, out_size, file_pcmdump );
#endif

	return out_size;

}
#endif

static int ffmpeg_dcatopcm( short * pcm_out, unsigned char * dca_in, int dca_in_size, int big_endian )
{
	int len;
	int tmp_size, out_size;
	short * pcm_ptr;

    avpacket.data = dca_in;
    avpacket.size = dca_in_size;

	pcm_ptr = pcm_out;
	out_size = 0;
	while( avpacket.size > 0 ){
	    DCAContext *s = (DCAContext *)pCodecCtx->priv_data;

		tmp_size = AVCODEC_MAX_AUDIO_FRAME_SIZE;
		len = ffmpeg.avcodec_decode_audio3(pCodecCtx, pcm_ptr, &tmp_size, &avpacket);
		if( len < 0 ){
			LOG("Error while decoding");
			PRINTHEX("DCA Frame data/32:", avpacket.data, 32 );
			return -1;
		}
// 		PRINTF("frame_size = %i", s->frame_size );

		/* PCM_S16BE "encoding" -> change from LE16 to BE16 */
		if( big_endian )
			change_endian16( pcm_ptr, pcm_ptr, tmp_size >> 1 );

		out_size      += tmp_size;
		pcm_ptr       += (tmp_size >> 1);
		avpacket.data += s->frame_size;
		avpacket.size -= s->frame_size;

		/* increment number of samples */
		total_samples += (tmp_size >> 1);

	}

/*
	if( (pCodecCtx->sample_rate != 48000) ||
		(pCodecCtx->bit_rate != 1536000) ) {
		PRINTF("Decoded DCA packet needs resampling (pCodecCtx->bit_rate=%i, pCodecCtx->sample_rate=%i)", pCodecCtx->bit_rate, pCodecCtx->sample_rate );
	
		pcm_ptr = pcm_out;
//		ffmpeg.audio_resample( resample, (short *)framebuf, pcm_out, out_size>>1 );
//		pcm_ptr = (short *)framebuf;
	} else {
		pcm_ptr = pcm_out;
	}
*/

#if DUMP_RAWPCM
	fwrite( (void *)pcm_out, 1, out_size, file_pcmdump );
#endif

	return out_size;
}

/* The extension to the original spIAe_Open function */
int spIAe_Open_DTS( struct avfd_instance_t * avfd , void * result )
{

	PRINTF("%s In", __FUNCTION__);

	/* Initialize ffmpeg decoder, if required */
	if( soft_codec_id != CODEC_ID_NONE ){
	
		ffmpeg.av_init_packet(&avpacket);

		/* find the requested audio decoder */
		pCodec = ffmpeg.avcodec_find_decoder(soft_codec_id);
		if( !pCodec ){
			LOG("Error: required codec (0x%X) not found. Please check your ffmpeg library.", soft_codec_id);
			return 0x8522DFF9;
		}
	    pCodecCtx = ffmpeg.avcodec_alloc_context();

		/* enable stereo downmix */
		pCodecCtx->request_channels = 2;

		/* open it */
		if( ffmpeg.avcodec_open(pCodecCtx, pCodec) < 0 ){
			LOG("Error: could not open codec");
			return 0x8522DFF9;
		}

		/* TODO: create the resample context(s) */
		/* needed as soon as ffmpeg supports X96 extension */
//		resample = ffmpeg.av_audio_resample_init( 2, 6, 48000, 48000,
//			SAMPLE_FMT_S16, SAMPLE_FMT_S16, 16, 10, 0, 0.8);
//		if( !resample ){
//			LOG("Error: could not create resample context.");
//			return 0x8522DFF9;
//		}

		/* framebuf contains audio data from feeder */
	    framebuf = (unsigned char *)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
		/* samples will contain decoded audio samples @ 48000 Hz */
		samples = (short *)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);

		/* just for debugging, a sample counter */
		total_samples = 0;

#if DUMP_RAWPCM
		file_pcmdump = fopen( RAWPCM_DUMPFILE, "wb" );
#endif

	}

	/* Continue with original spIAe_Open() function */
	return spIAe_Open( avfd, result );
}


/* The extension to the original spIAe_Close function */
int spIAe_Close_DTS( struct avfd_instance_t * avfd )
{
	PRINTF("%s In", __FUNCTION__);
	
	/* TODO: close the resample context(s) */
//	ffmpeg.audio_resample_close( resample );
	if( framebuf )
		free( framebuf );
	framebuf = NULL;
	if( samples )
		free( samples );
    samples = NULL;
	if( pCodecCtx ){
		ffmpeg.avcodec_close( pCodecCtx );
	    free( pCodecCtx );
		pCodecCtx = NULL;
	}

#if DUMP_RAWPCM
	fclose( file_pcmdump );
#endif

	/* Continue with original spIAe_Close() function */
	return spIAe_Close( avfd );
}

/* 009F6338 */
static int staAvfdIsDemuxStop_DTS( struct avfd_instance_t * avfd, int mode )
{
	if( spI_CheckPrintLevel(0x200000) )
		PRINTF("<+%s>", __FUNCTION__);

	if( (avfd->StreamHandle) ||
		(avfd->mRes382D8) ){
		if( spI_CheckPrintLevel(0x200000) )
			PRINTF("<-%s>", __FUNCTION__);
		return uldAvfd_InterfacerContainerIsDemuxStopped( avfd );
	} else {
		if( mode == 0 ){
			if( avfd->mRes38360 == 1 ){
				if( spI_CheckPrintLevel(0x200000) )
					PRINTF("<-%s>", __FUNCTION__);
				return 1;
			} else {
				if( spI_CheckPrintLevel(0x200000) )
					PRINTF("<-%s>", __FUNCTION__);
				return 0;
			}
		} else {
			if( avfd->mRes38364 == 1 ){
				if( spI_CheckPrintLevel(0x200000) )
					PRINTF("<-%s>", __FUNCTION__);
				return 1;
			} else {
				if( spI_CheckPrintLevel(0x200000) )
					PRINTF("<-%s>", __FUNCTION__);
				return 0;
			}
		}
	}
	return 0;
}

static int staAvfd_AeCopyData( struct avfd_instance_t * avfd, unsigned char * buffer, unsigned int datasize )
{
	unsigned int buffer_free_size;
	int bytes_to_write;
	unsigned char * read_ptr = buffer;

	// PRINTF("%s In", __FUNCTION__);

	if( spI_CheckPrintLevel(0x200000) )
		PRINTF("<+%s>", __FUNCTION__);

	if( datasize == 0 ){
		if( spI_CheckPrintLevel(0x200000) )
			PRINTF("<-%s>", __FUNCTION__);
		return 0;
	}
	//PRINTF("AE buff_start: 0x%08X, writeptr: 0x%08X, buff_end: 0x%08X, buff_offset: 0x%08X",
	//	(int)avfd->AudioBufferBase, (int)avfd->AudioWritePtr, (int)avfd->AudioBufferEnd , avfd->AudioBufOffset );
	while( 1 ){
		buffer_free_size = 0;
		staAvfd_GetAeBufferFreeSize(avfd, &buffer_free_size);
		if( datasize < buffer_free_size )
			bytes_to_write = datasize;
		else
			bytes_to_write = buffer_free_size;

		// PRINTF("bytes_to_write = 0x%04X", bytes_to_write);

		/* circular buffer handling */
		if( (avfd->AudioWritePtr + bytes_to_write) < avfd->AudioBufferEnd ){
			memcpy( avfd->AudioWritePtr, read_ptr, bytes_to_write );
			read_ptr += bytes_to_write;
			avfd->AudioWritePtr += bytes_to_write;
			avfd->avfd_interface->AudioCodecFunctions->SetWritePtr(
				avfd->avfd_interface->AvfdInterfaceData, (int)avfd->AudioWritePtr & ~1 );
		} else {
			int chunksize2;
			int chunksize1  = avfd->AudioBufferEnd - avfd->AudioWritePtr;
			chunksize1 -= avfd->AudioBufOffset;
			chunksize2 = bytes_to_write - chunksize1;
			memcpy( avfd->AudioWritePtr, read_ptr, chunksize1 );
			read_ptr += chunksize1;
			avfd->AudioWritePtr = avfd->AudioBufferBase;
			avfd->avfd_interface->AudioCodecFunctions->SetWritePtr(
				avfd->avfd_interface->AvfdInterfaceData, (int)avfd->AudioBufferBase+1 );
			memcpy( avfd->AudioBufferBase, read_ptr, chunksize2 );
			read_ptr += chunksize2;
			if( chunksize2 ){
				avfd->AudioWritePtr = avfd->AudioBufferBase + chunksize2;
				avfd->avfd_interface->AudioCodecFunctions->SetWritePtr(
					avfd->avfd_interface->AvfdInterfaceData, (int)avfd->AudioWritePtr & ~1 );
			}
		}
		datasize -= bytes_to_write;
		if( datasize == 0 ){
			if( spI_CheckPrintLevel(0x200000) )
				PRINTF("<-%s>", __FUNCTION__);
			// PRINTF("%s Out", __FUNCTION__);
			return 0x0;
		}
		if( staAvfdIsDemuxStop_DTS( avfd, 1 ) ){
			if( spI_CheckPrintLevel(0x200000) )
				PRINTF("<-%s>", __FUNCTION__);
			return 0xFFF;
		}
		OsaWait( 2 );
	}
} 

int InterfacerContainerAeCopyData_DTS(struct avfd_instance_t * avfd, unsigned char * buffer, int buffsize)
{
	//PRINTF("\nInterfacerContainerAeCopyData_DTS(0x%08X, 0x%08X, 0x%02X)", (int)avfd, int(buffer), buffsize); 
	//PRINTF("\nsoft_codec_id = 0x%X", soft_codec_id ); 
	//PRINTHEX( "audio data:", buffer, buffsize );
	
	/* when audio processing is done through AE */
	if( soft_codec_id == CODEC_ID_NONE )
		return uldAvfd_InterfacerContainerAeCopyData(  avfd, buffer, buffsize );

	/* just in case we received 2nd header block,
	   so pass 1st block to AE */
	if( audio_header_size ){
		int retval;
		LOG("Warning: received 2nd header block");
		retval = uldAvfd_InterfacerContainerAeCopyData(  avfd, audio_header, audio_header_size );
		audio_header_size = 0;
		return retval;
	}

	if( (unsigned)buffsize < sizeof(audio_header) ){
		memcpy( audio_header, buffer, buffsize );
		audio_header_size = buffsize;
		return 0;
	} else {
		LOG("Error: buffer too small to hold header data.");
		return 0;
	}
}

int InterfacerAudioTransferData_DTS(struct avfd_instance_t * avfd, int buffsize, int * param3 )
{
	int len;
	int out_size;
	int num_frames;
	
	//PRINTF("\nInterfacerAudioTransferData_DTS(0x%08X, 0x%08X, 0x%08X/0x%08X)", (int)avfd, int(buffsize), param3 ? (param3[0]):0, param3 ? (param3[1]):0 );
	//PRINTF("soft_codec_id = 0x%08X", soft_codec_id);

	/* when there is no need for software decoding,
	   just pass data to original funtion */
	if( soft_codec_id == CODEC_ID_NONE )
		return uldAvfd_InterfacerAudioTransferData( avfd, buffsize, param3 );

	/* according to disassembly this case will never happen (param3 is always 0).
	   checked with T-CHLCIPDEUC2007.1 */
	if( param3 ){
		if( (param3[1]) &&
			(param3[0]) ){
			PRINTHEX( "data (1):", (unsigned char *)&param3[0], param3[1] );
			uldAvfd_InterfacerContainerAeCopyData( avfd, (unsigned char *)&param3[0], param3[1] );
		}
	}

	if( framebuf == NULL ){
		PRINTF("[%s] error: framebuf is NULL!", __FUNCTION__ );
		return 0;
	}

	/* read and ignore audio substream header */
	avfd_GetBuf(avfd->StreamHandle, framebuf, 3 );
	buffsize -= 3;
	num_frames = framebuf[0];

	/* read DTS frame (should start with DTS sync word 7FFE8001) */
	avfd_GetBuf(avfd->StreamHandle, framebuf, buffsize );
	//PRINTF("DCA size = 0x%04X", buffsize );
	//PRINTHEX( "DCA data/64:", framebuf, 64 );

#ifdef __HAVE_LIBDCA__
	if( samygodca_use_libdca )
		out_size = libdca_dcatopcm( samples, framebuf, buffsize, 1 );
	else
#endif
		out_size = ffmpeg_dcatopcm( samples, framebuf, buffsize, 1 );

	if( audio_header_size && (out_size > 0) ){

		unsigned char lpcm_header[6];
		unsigned char audio_index;

		// de-mask substream number	
		audio_index = audio_header[audio_header_size - 1] & 7;
		if( audio_header[14] == 0xFF ){
			audio_header[8] = 0x06;
		} else {
			audio_header[8]  = 0x09;
			audio_header[14] = 0x10;
			audio_header[15] = 0x60;
			audio_header[16] = 0x08;	// org: 0x04
			audio_header[17] = 0xFF;
		}

		audio_header_size = audio_header[8]+10;

#if 1
		/* adjust len according to payload */
		len = out_size + audio_header_size; // + 6;
#else
		/* adjust payload according to len */
		len = 0x07F2;
		out_size = len - audio_header_size - 6;
		out_size &= 0xFFE;
		len = out_size + audio_header_size + 6;
#endif
		//out_size = buffsize+3;	// len - audio_header[8] - 4;
		audio_header[4] = (unsigned char)((len >> 8) & 0xFF);
		audio_header[5] = (unsigned char)((len) & 0xFF);
		/* insert new substream number or 0xA0 (LPCM substream number 0)*/
		audio_header[audio_header_size-1] = 0xA0 | audio_index;
		lpcm_header[0] = 0x07;	/* number of frames */
		lpcm_header[1] = 0x00;	/* 0004 */
		lpcm_header[2] = 0x04;	/* 0004 -> PTS frame @ offset 26 */
		lpcm_header[3] = 0x0C;	/* emphasis off (128), mute off (64), frame modulo 20 = 0x0C */
		lpcm_header[4] = 0x01;	/* 16-bit, 48K sample rate, stereo */
		lpcm_header[5] = 0x80;	/* dynamic range 128 (neutral) */
		//PRINTF("new audio packet: packet len = 0x%04X / data size = 0x%04X", len, out_size );
		//PRINTHEX("audio header:", audio_header, audio_header_size );
		//PRINTHEX("lpcm header:", lpcm_header, 6 );
		//PRINTHEX("audio data/64:", (unsigned char *)samples, 256 );
		PRINTF("total_samples = %i", total_samples );

		uldAvfd_InterfacerContainerAeCopyData( avfd, audio_header, audio_header_size );
		memcpy( framebuf, lpcm_header, sizeof(lpcm_header) );
		memcpy( framebuf+sizeof(lpcm_header), (void *)samples, out_size );
		staAvfd_AeCopyData( avfd, framebuf, out_size + sizeof(lpcm_header) );
		audio_header_size = 0;
		return 0;
	} else {
		return 0; // return uldAvfd_InterfacerContainerAeCopyData( avfd, trans_buf, buffsize );
	}
}

int InterfacerAudioTransferDataWithSyncHdr_DTS( struct avfd_instance_t * avfd, int param2, int param3, int param4, int param5)
{
	int out_size;
	unsigned char temp_header[sizeof(audiotransfer_sync_header)];


	if( spI_CheckPrintLevel(0x200000) )
		PRINTF("%s: p2=0x%08X, p3=0x%08X, p4=0x%08X, p5=0x%08X", __FUNCTION__, param2, param3, param4, param5 );

	/* when there is no need for software decoding,
	   just pass data to original funtion */
	if( soft_codec_id == CODEC_ID_NONE )
		return uldAvfd_InterfacerAudioTransferDataWithSyncHdr( avfd, param2, param3, param4, param5);

	if( framebuf == NULL ){
		PRINTF("[%s] error: framebuf is NULL!", __FUNCTION__ );
		return 0;
	}

	/* Quick hack for adjusting PTS */
	if( param3 == -1 ){
		last_pts += 960;	/* DCA frame size is 960 ticks of 90kHz clock long */
		param3 = last_pts;
	} else {
		last_pts = param3;
	}

	avfd_GetBuf(avfd->StreamHandle, framebuf+4, param2 );
//	PRINTF("DCA size = 0x%04X", param2 );
//	PRINTHEX( "DCA data/64:", framebuf, 64 );

#ifdef __HAVE_LIBDCA__
	if( samygodca_use_libdca )
		out_size = libdca_dcatopcm( samples, framebuf+4, param2, 0 );
	else
#endif
		out_size = ffmpeg_dcatopcm( samples, framebuf+4, param2, 0 );

	/* Quick hack: prepend DCA sync word before frame when decoding failed and try again */
	/*             this should cope with the "header removal compression" */
	if( out_size == -1 ){
		/* since framebuf is malloc'd, it will be algined at 16 byte boundary.
		   so this will not cause an invalid assignment exception */
		*(unsigned int *)framebuf = 0x0180FE7F; /* DCA_MARKER_RAW_BE = 0x7FFE8001 */
#ifdef __HAVE_LIBDCA__
		if( samygodca_use_libdca )
			out_size = libdca_dcatopcm( samples, framebuf, param2+4, 0 );
		else
#endif
			out_size = ffmpeg_dcatopcm( samples, framebuf, param2+4, 0 );
	}

	if( out_size > 0 ){

		memcpy( temp_header, audiotransfer_sync_header, sizeof(audiotransfer_sync_header) );

		if( param3 == -1 ){
			temp_header[0x08] = 0xFF;
			temp_header[0x09] = 0xFF;
			temp_header[0x0A] = 0xFF;
			temp_header[0x0B] = 0xFF;
		} else {
			temp_header[0x08] = (unsigned char)(param3 >> 25);
			temp_header[0x09] = (unsigned char)(param3 >> 17);
			temp_header[0x0A] = (unsigned char)(param3 >> 9);
			temp_header[0x0B] = (unsigned char)(param3 >> 1);
		}
		temp_header[0x0C] = (unsigned char)(out_size >> 24);
		temp_header[0x0D] = (unsigned char)(out_size >> 16);
		temp_header[0x0E] = (unsigned char)(out_size >> 8);
		temp_header[0x0F] = (unsigned char)(out_size);
		if( param5 == 0 )
			uldAvfd_InterfacerContainerAeCopyData( avfd, temp_header, sizeof(temp_header) );

//		PRINTF("LPCM packet: data size = 0x%04X", out_size );
//		PRINTHEX("audio data/64:", (unsigned char *)samples, 256 );
//		PRINTF("total_samples = %i", total_samples );
		return staAvfd_AeCopyData( avfd, (unsigned char *)samples, out_size);
	}
	return 0;
}

int InterfacerAudioTransferBufDataWithSyncHdr_DTS( struct avfd_instance_t * avfd, unsigned char * param2, int param3, int param4)
{
	PRINTF("%s: p2=0x%08X, p3=0x%08X, p4=0x%08X", __FUNCTION__, (int)param2, param3, param4 );
	return uldAvfd_InterfacerAudioTransferBufDataWithSyncHdr( avfd, param2, param3, param4 );
}

/* 009F3AFC */
int uldAvfd_InterfacerGetAudioCodecType_DTS(struct avfd_instance_t * avfd, int audio_codec_id )
{
	PRINTF("%s: audio_codec_id = %i", __FUNCTION__, audio_codec_id );
	switch( audio_codec_id ){
	case 0:
		return 0;
	case 1:
		return 1;
	case 2:
	case 7:
		return 2;
	case 3:
		return 3;
	case 4:
		return 4;
	case 5:
		return 5;
	case 6:
		return 6;
	case 8:
		return 7;
	case 10:
		PRINTF("[%s] got DTS/DCA codec. Returning PCM codec id.", __FUNCTION__ );
		return 2;
	default:
		return 8;
	}
}

int staMkv_GetAudioCodecType_DTS( struct mkv_frame_t * mkv_frame )
{
	int retval;
	unsigned int codec_param = mkv_frame->codec_param;
	char * codec_name = mkv_frame->codec_name;

	if( spI_CheckPrintLevel(0x200000) )
		PRINTF("[%s] In", __FUNCTION__);

	/* default */
	retval = 9;
	for( ; ; ){
		if( !strcmp(codec_name, "A_AAC") ||
			!strcmp(codec_name, "A_AAC/MPEG2/LC") ||
			!strcmp(codec_name, "A_AAC/MPEG2/LC/SBR") ||
			!strcmp(codec_name, "A_AAC/MPEG4/LC") ||
			!strcmp(codec_name, "A_AAC/MPEG4/LC/SBR") ){
			retval = 6;
			break;
		}
		if( !strcmp(codec_name, "A_AC3") ){
			retval = 4;
			break;
		}
		if( !strcmp(codec_name, "A_DTS") ){
//			retval = 9;
			retval = 10;	/* unused ID for DTS */
			break;
		}
		if( !strcmp(codec_name, "A_MPEG/L2") ||
			!strcmp(codec_name, "A_MPEG/L1") ){
			retval = 0;
			break;
		}
		if( !strcmp(codec_name, "A_MPEG/L3") ){
			retval = 3;
			break;
		}
		if( !strcmp(codec_name, "A_PCM/INT/LIT") ){
			retval = 2;
			break;
		}
		if( !strcmp(codec_name, "A_MS/ACM") ){
			switch( codec_param ){
			case 0x1:
				retval = 2;
				break;
			case 0x2:
				retval = 1;
				break;
			case 0x6:
			case 0x7:
				retval = 8;
				break;
			case 0x11:
				retval = 1;
				break;
			case 0x50:
			case 0x55:
				retval = 3;
				break;
			case 0x161:
			case 0x162:
			case 0x163:
				retval = 5;
				break;
			case 0x2000:
				retval = 4;
				break;
			}
			break;
		}
		break;
	}

	if( spI_CheckPrintLevel(0x200000) )
		PRINTF("[%s] Out", __FUNCTION__);

	return retval;
}

extern "C" void uldAvfd_MkvDemux( struct avfd_instance_t * avfd );

void uldAvfd_MkvDemux_DTS( struct avfd_instance_t * avfd )
{
	PRINTF("[%s] In", __FUNCTION__ );
	uldAvfd_MkvDemux( avfd );
	PRINTF("[%s] Out", __FUNCTION__ );
}


int samygodca_init(const char *ffmpeg_path)
{
	ffmpeg.initialised = 0;

	PRINTF("%s: In", __FUNCTION__ );

	/* set path for external ffmpeg library */
	strcpy( path_ext_ffmpeg, ffmpeg_path );

	/* load ffmpeg library and make sure imports are loaded */
	CParsingMedia * mediaparser;
	mediaparser = new(CParsingMedia);
	mediaparser->PrepareParsingExtFfmpeg();
	mediaparser->LoadImportsFfmpeg( &ffmpeg );
	delete( mediaparser );

	/* register libdca state */
#ifdef __HAVE_LIBDCA__
	libdca.initialised = 0;
	PrepareLibDCA( &libdca, ffmpeg_path );

	/* if libdca could be initialized -> use it */
	if( libdca.initialised ){
		PRINTF("Now using libdca for DCA decoding.");
		libdca_state = libdca.dca_init(0);
	} else {
		samygodca_use_libdca = 0; 
#else
	{
#endif
		if( ffmpeg.initialised ){
			ffmpeg.avcodec_register_all();
			ffmpeg.av_register_all(); 
			PRINTF("Now using ffmpeg for DCA decoding.");
		} else {
			LOG("libdca and ffmpeg are not found");
			return (EX_SOFTWARE);
		}
	}

	framebuf = NULL;
	samples = NULL;
	pCodecCtx = NULL;

	PRINTF("%s: Out", __FUNCTION__ );

	return (EX_OK);
}
