/*
 * samygodca_inj.hpp
 * Copyright (C) 2010 Andreas Fiedler <smartsmurf@users.sourceforge.net>
 *
 * This file is part of samygodca, a free DTS Coherent Acoustics
 * stream decoder plugin for SamyGO extension of Samsung DTV.
 *
 * samygodca is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * samygodca is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef __SAMYGODCA_INJ_HPP__
#define __SAMYGODCA_INJ_HPP__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <dlfcn.h>
#include <sys/mman.h>

#include "CParsingMedia.hpp"
#include "CMovieCore.hpp"
#include "uldAvfdInterfacer.h"
#include "spIAe.h"
#include "uldAvfdMkvControl.h"

/* ffmpeg includes */
#include <libavcodec/avcodec.h>
#include <libavcodec/get_bits.h>


/* taken from dca.c */
#define DCA_PRIM_CHANNELS_MAX (7)
#define DCA_SUBBANDS (32)
#define DCA_ABITS_MAX (32)      /* Should be 28 */
#define DCA_SUBSUBFRAMES_MAX (4)
#define DCA_SUBFRAMES_MAX (16)
#define DCA_BLOCKS_MAX (16)
#define DCA_LFE_MAX (3)

enum DCAMode {
    DCA_MONO = 0,
    DCA_CHANNEL,
    DCA_STEREO,
    DCA_STEREO_SUMDIFF,
    DCA_STEREO_TOTAL,
    DCA_3F,
    DCA_2F1R,
    DCA_3F1R,
    DCA_2F2R,
    DCA_3F2R,
    DCA_4F2R
}; 

#define DCA_MAX_FRAME_SIZE 16384 

typedef struct {
    AVCodecContext *avctx;
    /* Frame header */
    int frame_type;             ///< type of the current frame
    int samples_deficit;        ///< deficit sample count
    int crc_present;            ///< crc is present in the bitstream
    int sample_blocks;          ///< number of PCM sample blocks
    int frame_size;             ///< primary frame byte size
    int amode;                  ///< audio channels arrangement
    int sample_rate;            ///< audio sampling rate
    int bit_rate;               ///< transmission bit rate
    int bit_rate_index;         ///< transmission bit rate index

    int downmix;                ///< embedded downmix enabled
    int dynrange;               ///< embedded dynamic range flag
    int timestamp;              ///< embedded time stamp flag
    int aux_data;               ///< auxiliary data flag
    int hdcd;                   ///< source material is mastered in HDCD
    int ext_descr;              ///< extension audio descriptor flag
    int ext_coding;             ///< extended coding flag
    int aspf;                   ///< audio sync word insertion flag
    int lfe;                    ///< low frequency effects flag
    int predictor_history;      ///< predictor history flag
    int header_crc;             ///< header crc check bytes
    int multirate_inter;        ///< multirate interpolator switch
    int version;                ///< encoder software revision
    int copy_history;           ///< copy history
    int source_pcm_res;         ///< source pcm resolution
    int front_sum;              ///< front sum/difference flag
    int surround_sum;           ///< surround sum/difference flag
    int dialog_norm;            ///< dialog normalisation parameter

    /* Primary audio coding header */
    int subframes;              ///< number of subframes
    int total_channels;         ///< number of channels including extensions
    int prim_channels;          ///< number of primary audio channels
    int subband_activity[DCA_PRIM_CHANNELS_MAX];    ///< subband activity count
    int vq_start_subband[DCA_PRIM_CHANNELS_MAX];    ///< high frequency vq start subband
    int joint_intensity[DCA_PRIM_CHANNELS_MAX];     ///< joint intensity coding index
    int transient_huffman[DCA_PRIM_CHANNELS_MAX];   ///< transient mode code book
    int scalefactor_huffman[DCA_PRIM_CHANNELS_MAX]; ///< scale factor code book
    int bitalloc_huffman[DCA_PRIM_CHANNELS_MAX];    ///< bit allocation quantizer select
    int quant_index_huffman[DCA_PRIM_CHANNELS_MAX][DCA_ABITS_MAX]; ///< quantization index codebook select
    float scalefactor_adj[DCA_PRIM_CHANNELS_MAX][DCA_ABITS_MAX];   ///< scale factor adjustment

    /* Primary audio coding side information */
    int subsubframes[DCA_SUBFRAMES_MAX];           ///< number of subsubframes
    int partial_samples[DCA_SUBFRAMES_MAX];        ///< partial subsubframe samples count
    int prediction_mode[DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS];    ///< prediction mode (ADPCM used or not)
    int prediction_vq[DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS];      ///< prediction VQ coefs
    int bitalloc[DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS];           ///< bit allocation index
    int transition_mode[DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS];    ///< transition mode (transients)
    int scale_factor[DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS][2];    ///< scale factors (2 if transient)
    int joint_huff[DCA_PRIM_CHANNELS_MAX];                       ///< joint subband scale factors codebook
    int joint_scale_factor[DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS]; ///< joint subband scale factors
    int downmix_coef[DCA_PRIM_CHANNELS_MAX][2];                  ///< stereo downmix coefficients
    int dynrange_coef;                                           ///< dynamic range coefficient

    int high_freq_vq[DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS];       ///< VQ encoded high frequency subbands

    float lfe_data[2 * DCA_LFE_MAX * (DCA_BLOCKS_MAX + 4)];      ///< Low frequency effect data
    int lfe_scale_factor;

    /* Subband samples history (for ADPCM) */
    float subband_samples_hist[DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS][4];
    DECLARE_ALIGNED(16, float, subband_fir_hist)[DCA_PRIM_CHANNELS_MAX][512];
    DECLARE_ALIGNED(16, float, subband_fir_noidea)[DCA_PRIM_CHANNELS_MAX][32];
    int hist_index[DCA_PRIM_CHANNELS_MAX];
    DECLARE_ALIGNED(16, float, raXin)[32];

    int output;                 ///< type of output
    float add_bias;             ///< output bias
    float scale_bias;           ///< output scale

    DECLARE_ALIGNED(16, float, subband_samples)[DCA_BLOCKS_MAX][DCA_PRIM_CHANNELS_MAX][DCA_SUBBANDS][8];
    DECLARE_ALIGNED(16, float, samples)[(DCA_PRIM_CHANNELS_MAX+1)*256];
    const float *samples_chanptr[DCA_PRIM_CHANNELS_MAX+1];

    uint8_t dca_buffer[DCA_MAX_FRAME_SIZE];
    int dca_buffer_size;        ///< how much data is in the dca_buffer

    const int8_t* channel_order_tab;                             ///< channel reordering table, lfe and non lfe
    GetBitContext gb;
    /* Current position in DCA frame */
    int current_subframe;
    int current_subsubframe;

    /* XCh extension information */
    int xch_present;
    int xch_base_channel;       ///< index of first (only) channel containing XCH data

    int debug_flag;             ///< used for suppressing repeated error messages output
    //DSPContext dsp;
    //FFTContext imdct;
    //SynthFilterContext synth;
    //DCADSPContext dcadsp;
} DCAContext; 

/* libdca includes */
#ifdef __HAVE_LIBDCA__
#include <dca.h>

/* libdca library interface */
struct libdca_interface_t {
	dca_state_t * (*dca_init)(uint32_t mm_accel);
	int (*dca_syncinfo)(dca_state_t *state, uint8_t * buf, int * flags,
						int * sample_rate, int * bit_rate, int *frame_length);
	int (*dca_frame)(dca_state_t * state, uint8_t * buf, int * flags,
					level_t * level, sample_t bias);
	void (*dca_dynrng)(dca_state_t * state,
					level_t (* call) (level_t, void *), void * data);
	int (*dca_blocks_num)(dca_state_t * state);
	int (*dca_block)(dca_state_t * state);
	sample_t * (*dca_samples)(dca_state_t * state);
	void (*dca_free)(dca_state_t * state);
	void * hLib_libdca;
	int  initialised;
};
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __HAVE_LIBDCA__
extern struct libdca_interface_t libdca;
static void PrepareLibDCA( struct libdca_interface_t * dca, const char * lib_path );
#endif

/* media parsing and handling functions */
extern void CParsingMedia_PrepareParsing(CParsingMedia * Instance);
extern int CParsingMedia_GetMediaTypeInfo(CParsingMedia * Instance, char  const* file_name, Multimedia::_tMediaInfoContext * media_info_context );
extern int CheckSupportMovie_DTS( CMovieTDIF::MovieSettings );
//extern int CMovieCore_SelectAudio(CMovieCore *, Multimedia::_tMediaInfoContext * , unsigned int );
static void change_endian16( short * OutBuf, const short * InBuf, const int words );
static int staAvfd_AeCopyData( struct avfd_instance_t *, unsigned char *, unsigned int );

/* avfd audio interface functions */ 
extern int spIAe_Open_DTS( struct avfd_instance_t * avfd , void * result );
extern int spIAe_Close_DTS( struct avfd_instance_t * avfd );
extern int InterfacerContainerAeCopyData_DTS(struct avfd_instance_t * avfd, unsigned char * buffer, int buffsize);
extern int InterfacerAudioTransferData_DTS(struct avfd_instance_t * avfd, int buffsize, int * param3 );
extern int InterfacerAudioTransferDataWithSyncHdr_DTS( struct avfd_instance_t *, int, int, int, int );
extern int InterfacerAudioTransferBufDataWithSyncHdr_DTS( struct avfd_instance_t *, unsigned char *, int, int );

/* DTS to PCM transcoding functions */
#ifdef __HAVE_LIBDCA__
static int libdca_dcatopcm( short *, unsigned char *, int, int );
#endif
static int ffmpeg_dcatopcm( short *, unsigned char *, int, int );

extern int uldAvfd_InterfacerGetAudioCodecType_DTS(struct avfd_instance_t * avfd, int audio_codec_id );
//extern int modAe_SetStartingStc_DTS( void *, int );
//extern int uldAvfd_StoreAudioPTS_DTS( float , int64_t , float );
//extern int staMkv_DumpFrame_DTS( struct avfd_instance_t * , struct mkv_frame_t * , uint64_t , uint64_t );
extern int staMkv_GetAudioCodecType_DTS( struct mkv_frame_t * );
extern void uldAvfd_MkvDemux_DTS( struct avfd_instance_t * avfd );

extern int samygodca_init(const char *ffmpeg_path);

#ifdef __cplusplus
}
#endif

#endif 
