#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "mod_int.h"

#define FUNC_ACTION_BY_CODE(suff)					\
	void *core_action_ ## suff (uint32_t code) {	\
		switch (code) {								\
		case CFG_ACTION_FILL:						\
			return (core_fill_ ## suff);			\
		case CFG_ACTION_PRINT:						\
			return (core_print_ ## suff);			\
		default:									\
			return (NULL);							\
		}											\
	}

/* fill configuration variables with default values */
struct core_config config = {
	.debug = 1,
	.exedsp_debug = 0,
	.exedsp_facility = 0xffffffff,
	.exedsp_severity = 0x100000,
	.mod_key = KEY_GREEN,
	.time_adjust = 0,
	.cip = CIP_AUTO,
};

static struct cfg_enum_entry ctl_cip[] = {
	{"YES",   CIP_YES},
	{"1",     CIP_YES},
	{"NO",    CIP_NO},
	{"0",     CIP_NO},
	{"AUTO",  CIP_AUTO},
	{NULL, 0}
};

static const struct mod_config_var vars[] = {
	{"LOG",             0, FLAG_ENTRY(config.debug)},
	{"EXEDSP_LOG",      0, FLAG_ENTRY(config.exedsp_debug)},
	{"EXEDSP_FACILITY", 0, INT_ENTRY(config.exedsp_facility)},
	{"EXEDSP_SEVERITY", 0, INT_ENTRY(config.exedsp_severity)},
	{"MOD_KEY",         0, KEY_ENTRY(config.mod_key)},
	{"TIME_ADJUST",     0, SINT_ENTRY(config.time_adjust)},
	{"CIPLUS",          0, ENUM_ENTRY(config.cip, ctl_cip)},
	{LAST_VAR_ENTRY}
};

static int
core_fill_int(char *name, char *str, void *ptr, void *ctl)
{
	long long int var;

	if (sscanf(str, "%lli", &var) != 1)
		var = 0;

	if (var > 0xffffffff || var < 0)
		var = 0;
	*(unsigned *)ptr = var;
	
	return (EX_OK);
}

static int
core_print_int(char *buf, size_t bufsize, void *value, void *ctl)
{
	snprintf(buf, bufsize, "%u / 0x%X",
			 *(unsigned *)value, *(unsigned *)value);
	return (EX_OK);
}

static int
core_fill_sint(char *name, char *str, void *ptr, void *ctl)
{
	*(int *)ptr = atoi(str);
	
	return (EX_OK);
}

static int
core_print_sint(char *buf, size_t bufsize, void *value, void *ctl)
{
	snprintf(buf, bufsize, "%d", *(int *)value);
	return (EX_OK);
}

static int
core_fill_flag(char *name, char *str, void *ptr, void *ctl)
{
	int *flag = ptr;

	if (!strcasecmp(str, "yes") || !strcasecmp(str, "true") ||
		!strcasecmp(str, "enable") || !strcasecmp(str, "enabled") ||
		!strcasecmp(str, "on")) {

		*flag = 1;
	} else if (!strcasecmp(str, "no") || !strcasecmp(str, "false") ||
			   !strcasecmp(str, "disable") || !strcasecmp(str, "disabled") ||
			   !strcasecmp(str, "off")) {
		*flag = 0;
	} else if (*str == '\0')
		*flag = 0;
	else
		*flag = (atoi(str) > 0);

	return (EX_OK);
}

static int
core_print_flag(char *buf, size_t bufsize, void *value, void *ctl)
{
	int *flag = value;

	strlcpy(buf, (*flag) ? "true" : "false", bufsize);
	return (EX_OK);
}

static int
core_fill_string(char *name, char *str, void *ptr, void *ctl)
{
	size_t size = (size_t)ctl;

	strlcpy((char *)ptr, str, size);

	return (EX_OK);
}

static int
core_print_string(char *buf, size_t bufsize, void *value, void *ctl)
{
	strlcpy(buf, value, bufsize);
	return (EX_OK);
}

static int
core_fill_path(char *name, char *str, void *ptr, void *ctl)
{
	size_t size = (size_t)ctl;

	if (str[0] == '/')
		strlcpy((char *)ptr, str, size);
	else
		snprintf(ptr, size, "%s/%s", get_app_path(), str);
	return (EX_OK);
}

static int
core_print_path(char *buf, size_t bufsize, void *value, void *ctl)
{
	return (core_print_string(buf, bufsize, value, ctl));
}

static int
core_fill_enum(char *name, char *str, void *ptr, void *ctl)
{
	struct cfg_enum_entry *entry = ctl;

	for (; entry->name != NULL; entry++) {
		if (strcasecmp(str, entry->name))
			continue;
		*(int *)ptr = entry->value;
		return (EX_OK);
	}
	LOG("Unknown %s value[%s], leave the previous[%d / %X]",
		name, str, *(int *)ptr, *(int *)ptr);
	return (EX_SOFTWARE);
}

static int
core_print_enum(char *buf, size_t bufsize, void *value, void *ctl)
{
	struct cfg_enum_entry *entry = ctl;
	int val = *(int *)value;

	for (; entry->name != NULL; entry++) {
		if (val == entry->value) {
			snprintf(buf, bufsize, "%d/0x%X/%s", val, val, entry->name);
			return (EX_OK);
		}
	}
	strlcpy(buf, "*UNKNOWN*", bufsize);
	return (EX_SOFTWARE);
}

static int
core_fill_string_list(char *name, char *str, void *ptr, void *ctl_)
{
	struct stringlist *list = ptr;
	struct stringlist_entry *entry;
	struct cfg_stringlist_ctl *ctl = ctl_;
	char *substr;

	while ((entry = TAILQ_FIRST(list)) != NULL) {
		TAILQ_REMOVE(list, entry, link);
		free(entry);
	}

	substr = strtok(str, ctl->delim);
	for (; substr != NULL; substr = strtok(NULL, ctl->delim)) {
		if (ctl->check_func != NULL && ctl->check_func(substr) != EX_OK)
			continue;
		entry = calloc(1, sizeof(*entry));
		strlcpy(entry->value, substr, sizeof(entry->value));
		TAILQ_INSERT_TAIL(list, entry, link);
	}
#if 0
	LOG("%s =", name);
	TAILQ_FOREACH(entry, list, link)
		LOG("  [%s]", entry->value);
#endif
	return (0);
}

static int
core_print_string_list(char *buf, size_t bufsize, void *value, void *ctl_)
{
	struct cfg_stringlist_ctl *ctl = ctl_;
	struct stringlist *list = value;
	struct stringlist_entry *entry;
	bool first = true;
	int res;

	TAILQ_FOREACH(entry, list, link) {
		if (first) {
			res = snprintf(buf, bufsize, "%s", entry->value);
			first = false;
		} else
			res = snprintf(buf, bufsize, "%c%s", ctl->delim[0], entry->value);
		if (res >= bufsize)
			return (EX_OK);
		buf += res;
		bufsize -= res;
	}
	return (EX_OK);
}

static int
core_fill_key(char *name, char *str, void *ptr, void *ctl)
{
	char *str_orig = str;
	int key = 0;
	int rc = -1;
	struct {
		char ch;
		int flag;
		int present;
	} mods[] = {
		{'M', MKEY, 0},
		{'T', TKEY, 0},
		{0,   0,    0}
	}, *mod;

	if (!strcasecmp(str, "no") || !strcasecmp(str, "false") ||
		!strcasecmp(str, "disable") || !strcasecmp(str, "disabled") ||
		!strcasecmp(str, "off")) {

		rc = 0;
		goto DONE;
	}
    while (1) {
		if (str[0] != '\0' && str[1] == '+') {
			/* Try to parse modifier keys */
			char ch = toupper(str[0]);

			for (mod = mods; mod->ch != 0; mod++) {
				if (mod->present || ch != mod->ch)
					continue;
				/* ch == mod->ch */
				key |= mod->flag;
				mod->present = 1;
				str += 2;
				break;
			}
			if (mod->ch != 0)
				continue;
		}

		/* Try hexadecimal key value */
		if (str[0] == '0' && str[1] == 'x') {
			int tkey;
			int scanned;

			if (sscanf(str + 2, "%X%n", &tkey, &scanned) != 1)
				goto DONE;
			if (str[scanned + 2] != '\0')
				goto DONE;

			for (mod = mods; mod->ch != 0; mod++)
				if ((tkey & mod->flag))
					goto DONE;

			key |= tkey;
			rc = 0;
			goto DONE;
		} else {
			/* Try named key value */
			struct keyinfo *info;

			for (info = keyinfos; info->name != NULL; info++) {
				if (strcasecmp(info->name, str))
					continue;
				if (str[strlen(info->name)] != '\0')
					continue;
				key |= info->key;
				rc = 0;
				goto DONE;
			}
			goto DONE;
		}
		break;
	}

  DONE:
	if (rc != 0) {
		LOG("%s: cannot parse key[%s], leave the previous value[0x%X]",
			name, str_orig, *(int *)ptr);
	} else {
		*(int *)ptr = key;
	}
	return (EX_OK);
}

static int
core_print_key(char *buf, size_t bufsize, void *value, void *ctl_)
{
	return (key2str(buf, bufsize, *(int *)value));
}

#if 0
static int
set_default_config(void)
{
	struct stringlist_entry	*str_entry;
	
	/* TODO: move static cfgstruct structure initialization here */

	str_entry = calloc(1, sizeof(*str_entry));
	strlcpy(str_entry->value, "/dtv/usb/sda1", sizeof(str_entry->value));
	TAILQ_INSERT_TAIL(&config.pvr_target, str_entry, link);

	str_entry = calloc(1, sizeof(*str_entry));
	strlcpy(str_entry->value, "/mtd_wiselink/MOVIE001", sizeof(str_entry->value));
	TAILQ_INSERT_TAIL(&config.pvr_target, str_entry, link);

	return (0);
}
#endif

const struct mod_config_var *
mod_get_config_vars(void)
{
	return (vars);
}

FUNC_ACTION_BY_CODE(string);
FUNC_ACTION_BY_CODE(flag);
FUNC_ACTION_BY_CODE(int);
FUNC_ACTION_BY_CODE(sint);
FUNC_ACTION_BY_CODE(enum);
FUNC_ACTION_BY_CODE(string_list);
FUNC_ACTION_BY_CODE(key);
FUNC_ACTION_BY_CODE(path);
