/*
 * Energy-saving module
 */
#include "mod_esaving.h"

struct esaving_config config = {
	.key_check = MKEY|KEY_NUM0,
	.key_change = MKEY|KEY_NUM1,
	.key_mute = MKEY|KEY_NUM2,
	.key_panel_off = MKEY|KEY_NUM3,
};

static const struct mod_config_var config_vars[] = {
	{"ESAVING_CHECK",     0, KEY_ENTRY(config.key_check)},
	{"ESAVING_MUTE",      0, KEY_ENTRY(config.key_mute)},
	{"ESAVING_CHANGE",    0, KEY_ENTRY(config.key_change)},
	{"ESAVING_PANEL_OFF", 0, KEY_ENTRY(config.key_panel_off)},
	{LAST_VAR_ENTRY}
};

/* TCWindow::MuteVideo(bool) */
static int (*TCWindow_MuteVideo)(void **AppWindow, int flag);
#define TCTV_OPTION_ESAVING 0xf
/* TCTv::GetOption(int, int *) */
static int (*TCTv_GetOption)(void *AppTV, int option, unsigned *rval);
/* TCTv::SetOption(int, int) */
static int (*TCTv_SetOption)(void *AppTV, int option, int val);
/* CApp::SendEvent(PTEvent *, DTV_APPLICATION, bool) - CI version */
static int (*CApp_SendEvent_nocip)(void *TEvent, int what, int flag);
/* CApp::SendEvent(long, DTV_APPLICATION, bool) - CI+ version */
static int (*CApp_SendEvent_cip)(long type, int what, int flag);

static void *g_AppTV;
static void *g_AppWindow;
static void *g_AppWindowPIP;

static int cur_saving_mode; /* Initializing in esaving_init() */
static int panel_off = 0;

static void
esaving_mute(int key)
{
	static int mflag = 1;

	TCWindow_MuteVideo(g_AppWindow, mflag);
	TCWindow_MuteVideo(g_AppWindowPIP, mflag);

	LOG("Video mute %s", mflag ? "enabled" : "disabled");

	mflag = !mflag;
}

static unsigned int
esaving_get_option(void)
{
	unsigned int varr[2];
	unsigned int *rval=varr;
	int retv;

	memset(varr, 0, sizeof(varr));
	retv = TCTv_GetOption(g_AppTV, TCTV_OPTION_ESAVING, rval);

	/*
     * return: 1 on success
     * Off: 0, Low: 1, Medium: 2, High: 3, Auto: 4, Off: (0)*** 
     */
	LOG("TCTv::GetOption(0xf): ret=0x%X, value=0x%X", retv, *rval);

	return (*rval);
}

static void
esaving_check_mode(int key)
{
	esaving_get_option();
}

static int
esaving_set_option(int mode)
{
	int retv;

	retv = TCTv_SetOption(g_AppTV, TCTV_OPTION_ESAVING, mode);
	LOG("TCTv::SetOption(OPTION_ESAVING): esaving[0x%X] ret[0x%X]",
		mode, retv);
	return (retv);
}

static void
esaving_change_mode(int key)
{
	if (++cur_saving_mode > 4)
		cur_saving_mode = 0;
	esaving_set_option(cur_saving_mode);
	panel_off = 0;
}

static void
esaving_panel_off(int key)
{
	if (!panel_off) {
		TCWindow_MuteVideo(g_AppWindow, 1);
		TCWindow_MuteVideo(g_AppWindowPIP, 1);

		if (esaving_set_option(5)) {
			if (exedsp_is_cip())
				CApp_SendEvent_cip(0x434, 2, 0);
			else {
				int ptevent[6];

				memset(ptevent, 0, sizeof(ptevent));
				ptevent[3] = 0x432;
				ptevent[4] = 1;
				ptevent[5] = 0;
				CApp_SendEvent_nocip(ptevent + 1, 2, 0);
			}
		}
		panel_off = 1;
	} else {
		/* return to the previous saving mode */
		TCWindow_MuteVideo(g_AppWindow, 0);
		TCWindow_MuteVideo(g_AppWindowPIP, 0);

		esaving_set_option(cur_saving_mode);
		panel_off = 0;
	}
}

int
mod_keypressed(const int *keys, int keys_num, int *key_out)
{
	struct {
		int key;
		void (*func)(int key);
	} table[] = {
		{config.key_mute,      esaving_mute},
		{config.key_check,     esaving_check_mode},
		{config.key_change,    esaving_change_mode},
		{config.key_panel_off, esaving_panel_off},
		{0, NULL}
	}, *cur;

	for (cur = table; cur->func != NULL; cur++) {
		if (keys[0] != cur->key)
			continue;
		cur->func(keys[0]);
		return (MOD_KEYPRESSED_PROCESSED);
	}
	return (MOD_KEYPRESSED_UNKNOWN);
}

int
mod_init(void *handle)
{
	struct lib_symbol syms_nocip[] = {
		{SYMENTRY("_ZN8TCWindow9MuteVideoEb", &TCWindow_MuteVideo)},
		{SYMENTRY("_ZN4TCTv9GetOptionEiPi", &TCTv_GetOption)},
		{SYMENTRY("_ZN4TCTv9SetOptionEii", &TCTv_SetOption)},
		{SYMENTRY("_ZN4CApp9SendEventEP7PTEvent15DTV_APPLICATIONb", &CApp_SendEvent_nocip)},
		{SYMENTRY("g_AppTV", &g_AppTV)},
		{SYMENTRY("g_AppWindow", &g_AppWindow)},
		{SYMENTRY("g_AppWindowPIP", &g_AppWindowPIP)},
		{NULL, NULL}
	};
	struct lib_symbol syms_cip[] = {
		{SYMENTRY("_ZN8TCWindow9MuteVideoEb", &TCWindow_MuteVideo)},
		{SYMENTRY("_ZN4TCTv9GetOptionEiPi", &TCTv_GetOption)},
		{SYMENTRY("_ZN4TCTv9SetOptionEii", &TCTv_SetOption)},
		{SYMENTRY("_ZN4CApp9SendEventEl15DTV_APPLICATIONb", &CApp_SendEvent_cip)},
		{SYMENTRY("g_AppTV", &g_AppTV)},
		{SYMENTRY("g_AppWindow", &g_AppWindow)},
		{SYMENTRY("g_AppWindowPIP", &g_AppWindowPIP)},
		{NULL, NULL}
	};
	int rc;

	if (exedsp_is_cip())
		rc = exedsp_load_symbols(syms_cip, NULL);
	else
		rc = exedsp_load_symbols(syms_nocip, NULL);

	if (rc)
		return (EX_SOFTWARE);

	cur_saving_mode = esaving_get_option();
	return (EX_OK);
}

const struct mod_config_var *
mod_get_config_vars(void)
{
	return (config_vars);
}
