/*
 * EPG functions.
 */
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <time.h>

#include <mod.h>
#include "epg.h"
#include "mod_pvr.h"

static int EPG_initialized = 0;

static unsigned *g_AppTV;
static void (*TCChannel)(unsigned *channel);
static void (*TCProgram)(unsigned *program);
static int (*TCTv_GetSystemInfoTime)(void *AppTV, time_t *unixtime, int zero);
static int (*CTimerUtil_GetSITTime)(struct EPG_datetime *datetime, int zero);
static int (*CAppUtil_GetCurrentCh)(unsigned *channel, int zero);
static int (*TCTv_GetPresentProgram)(void *AppTV, unsigned *channel,
									 time_t *unixtime, unsigned *program);
static int (*TCProgram_GetTitle)(unsigned *program, unsigned short *namebuffer,
								 unsigned *namebuflen);

/*
 * Channel and program are constant instances initialized in
 * EPG_init() and so should be static.
 */
static unsigned channel[2];
static unsigned program[2];

#define PROGRAMBUFSIZE 128
#define NAMEBUFSIZE	256

static int
unicode_convert(unsigned short *unicode, int nchars, char *buf,
				unsigned bufsize)
{
	if (config.unicode == CFG_UNICODE_UTF8)
		return (unicode2utf8(unicode, nchars, buf, bufsize));
	else
		return (unicode2ascii(unicode, nchars, buf, bufsize,
							  config.unicode == CFG_UNICODE_ASCII_TRANSCODE));
}


void EPG_init(void)
{
	struct lib_symbol syms[] = {
		{SYMENTRY("_ZN9TCChannelC1Ev", &TCChannel)},
		{SYMENTRY("_ZN9TCProgramC1Ev", &TCProgram)},
		{SYMENTRY("_ZN4TCTv17GetSystemInfoTimeEPmi", &TCTv_GetSystemInfoTime)},
		{SYMENTRY("_ZN10CTimerUtil10GetSITTimeEPN6PCTime6PTTimeEi", &CTimerUtil_GetSITTime)},
		{SYMENTRY("_ZN8CAppUtil12GetCurrentChEP9TCChanneli", &CAppUtil_GetCurrentCh)},
		{SYMENTRY("_ZN4TCTv17GetPresentProgramEP9TCChanneljP9TCProgram", &TCTv_GetPresentProgram)},
		{SYMENTRY("_ZNK9TCProgram8GetTitleEPcPj", &TCProgram_GetTitle)},
		{SYMENTRY("g_AppTV", &g_AppTV)},
		{NULL, NULL}
	};

	if (EPG_initialized)
		return;

	EPG_initialized = 1;

	exedsp_load_symbols(syms, NULL);

	if (TCChannel)
		TCChannel(channel);
	if (TCProgram)
		TCProgram(program);
}

static time_t EPG_get_sysinfo_time(void)
{
	time_t ret = 0;

	if (TCTv_GetSystemInfoTime && g_AppTV) {
		if (!TCTv_GetSystemInfoTime(g_AppTV, &ret, 0)) {
			LOG("Cannot get the current time");
			return (0);
		}
	}

	return (ret);
}

static time_t EPG_get_SIT_time(void)
{
	struct EPG_datetime datetime;
	struct tm tm;
	time_t ret;

	bzero(&datetime, sizeof(datetime));
	if (CTimerUtil_GetSITTime) {
		if (!CTimerUtil_GetSITTime(&datetime, 0)) {
			LOG("Cannot get the current(SIT) time");
			return (0);
		}
	}

	bzero(&tm, sizeof(tm));
	tm.tm_sec = datetime.second;
	tm.tm_min = datetime.minute;
	tm.tm_hour = datetime.hour;
	tm.tm_mday = datetime.day;
	tm.tm_mon = datetime.month - 1;
	tm.tm_year = datetime.year - 1900;

	ret = mktime(&tm);

	return (ret);
}

static time_t
EPG_get_current_time(void)
{
	union {
		time_t arr[3];
		struct {
			time_t sys;
			time_t sit;
			time_t loc;
		} t;
	} u;
	time_t now;
	int i, sel = 0;

#define FJAN2010 1262340000

	bzero(&u, sizeof(u));
	u.t.sys = EPG_get_sysinfo_time();
	u.t.sit = EPG_get_SIT_time();
	u.t.loc = time(NULL);

	for (i = 1; i < sizeof(u.arr) / sizeof(u.arr[0]); i++)
		if (u.arr[i] > u.arr[sel])
			sel = i;
	if (u.arr[sel] < FJAN2010) {
		LOG("Cannot get the current time, sys[%u] sit[%u] loc[%u]",
			(unsigned)u.t.sys, (unsigned)u.t.sit, (unsigned)u.t.loc);
		return (0);
	}
	now = adjust_sytem_time(u.arr[sel]);
	LOG("Time: sys[%u] sit[%u] loc[%u], sel[%u] adj[%u]",
		(unsigned)u.t.sys, (unsigned)u.t.sit, (unsigned)u.t.loc,
		(unsigned)u.arr[sel], (unsigned)now);
	return (now);
}

static int EPG_get_current_channel(unsigned *channel)
{
	int ret = 0;

	if (CAppUtil_GetCurrentCh)
		ret = CAppUtil_GetCurrentCh(channel, 0);

	if (ret == 0)
		LOG("Cannot get the current channel");
	else
		LOG("Got the current channel[0x%X, 0x%X]", channel[0], channel[1]);

	return (ret);
}

static int EPG_get_present_program(time_t unixtime, unsigned *channel, unsigned *program)
{
	int ret = 0;

	if (TCTv_GetPresentProgram && g_AppTV)
 	ret = TCTv_GetPresentProgram(g_AppTV, channel, &unixtime, program);
	if (ret == 0)
		LOG("Cannot get the current program");
	else
		LOG("Got the current program[0x%X, 0x%X]", program[0], program[1]);

	return (ret);
}

static int EPG_get_program_title(unsigned *program, char *progname, size_t siz)
{
	unsigned short namebuffer[NAMEBUFSIZE];
	unsigned namebuflen = sizeof(namebuffer);
	int ret;

	namebuffer[0] = 0;
	if (TCProgram_GetTitle)
		ret = TCProgram_GetTitle(program, namebuffer, &namebuflen);

	if (ret == 0) {
		LOG("Cannot get the program title");
		return (0);
	}
	if (namebuflen > sizeof(namebuffer))
		namebuflen = sizeof(namebuffer);
	unicode_convert(namebuffer, namebuflen / sizeof(namebuffer[0]), progname, siz);
	return (ret);
}

void EPG_get_data(char *recpartname, size_t siz)
{
	time_t unixtime = 0;
	char progname[PROGRAMBUFSIZE] = "";

	*recpartname = '\0';
	EPG_init();

	unixtime = EPG_get_current_time();

	if (EPG_get_current_channel(channel) &&
		EPG_get_present_program(unixtime, channel, program)) {

		EPG_get_program_title(program, progname, sizeof(progname));
	}

	if (!unixtime)
		snprintf(recpartname, siz, "%s_", progname);
	else {
		struct tm *tm = localtime(&unixtime);

		snprintf(recpartname, siz, "%4d-%02d-%02d_%02d-%02d_%s_",
				 tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
				 tm->tm_hour, tm->tm_min, progname);
	}
}
